/*
 * RADIUS
 * Remote Authentication Dial In User Service
 *
 */

/*
 *
 * Portions Copyright 1997, 1998 Damien Miller <dmiller@vitnet.com.sg>
 *
 * Permission to use, copy, modify, and distribute this software 
 * for any purpose and without fee is hereby granted, provided 
 * that this copyright and permission notice appear on all copies 
 * and supporting documentation.
 *
 * Damien Miller makes no representations about the suitability 
 * of this software for any purpose.  It is provided "as is" 
 * without express or implied warranty.
 *
 */

/*
 *
 * Livingston Enterprises, Inc.
 * 6920 Koll Center Parkway
 * Pleasanton, CA   94566
 *
 * Copyright 1992 Livingston Enterprises, Inc.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose and without fee is hereby granted, provided that this
 * copyright and permission notice appear on all copies and supporting
 * documentation, the name of Livingston Enterprises, Inc. not be used
 * in advertising or publicity pertaining to distribution of the
 * program without specific prior permission, and notice be given
 * in supporting documentation that copying and distribution is by
 * permission of Livingston Enterprises, Inc.   
 *
 * Livingston Enterprises, Inc. makes no representations about
 * the suitability of this software for any purpose.  It is
 * provided "as is" without express or implied warranty.
 *
 */

/*
 * Modified from attrprint.c by Damien Miller <dmiller@vitnet.com.sg>
 * Patches/Bugfixes/Constructive criticism welcome.
 *
 * For maximum viewing pleasure, set your tabs to 3 spaces.
 *
 * See also: db_log.h and radius_log.def
 *
 * db_log() walks through the linked list of attributes and builds
 * an SQL INSERT query out of the attributes specified in db_log.h
 *
 * SendQueryToDatabase() does the business of passing the query to
 * the database server. It currently uses MySQL, but it should be
 * very easy to modify to use just about any other SQL backend.
 *
 */


static char          sccsid[] =
"@(#)@(#)attrprint.c	1.3 Copyright 1992 Livingston Enterprises Inc";

#include	<sys/types.h>
#include	<sys/socket.h>
#include	<sys/time.h>
#include	<netinet/in.h>

#include	<stdio.h>
#include	<netdb.h>
#include	<pwd.h>
#include	<time.h>
#include	<syslog.h>
#include	<ctype.h>
#include	<mysql.h>

#include	"radius.h"
#include	"protos.h"
#include	"db_log.h"
#include	"db_params.h"

void                 SendQueryToDatabase(char *query);

/*************************************************************************
 *
 *	Function: db_log
 *
 *	Purpose: Writes the accounting entry to an SQL database
 *
 *************************************************************************/

void                 db_log(VALUE_PAIR * pair)
{
	DICT_VALUE          *dval;
	char                 buffer[32];
	u_char              *ptr;
	int                  len,
	                     number;
	unsigned int         c;
	unsigned int         log_attrib;
	static char         *row_entries[SQL_NUM_ATTRIBS];
	static char         *query;
	size_t               str_end;

	/* Initialise the record with NULL pointers */
	for (c = 0; c < SQL_NUM_ATTRIBS; c++)
		row_entries[c] = NULL;

	/* Loop through the linked list of attribute pairs */
	while (pair != NULL_PAIR)
	{
		/* Is this an attribute that we are interested in? */
		log_attrib = 0;
		for (c = 0; c < SQL_NUM_ATTRIBS; c++)
		{
			if (!strncmp(pair->name, sql_log_pairs[c], strlen(sql_log_pairs[c])))
			{
				/* If so, flag it. */
				log_attrib = 1;
				break;
			}
		}

		/* At this point "c" contains the index into the record array */
		if (log_attrib)
		{
			switch (pair->type)
			{
				case ASCEND_SPECIAL_STRING:
					if (pair->attribute != ASCEND_NUMBER_SESSIONS)
					{
						row_entries[c] = malloc((pair->size * 2) + 4);
						sprintf(row_entries[c], "\'0x");
						ptr = (u_char *) pair->strvalue;
						len = pair->size;
						str_end = strlen(row_entries[c]);
						while (len)
						{
							sprintf(&(row_entries[c][str_end]), "%02x", *ptr);
							ptr++;
							len--;
							str_end += 2;
						}
						sprintf(&(row_entries[c][str_end]), "\'");
					}
					else
					{
						row_entries[c] = malloc((pair->size * 2) + 32);
						ptr = (u_char *) pair->strvalue + 4;

						number = ntohl(*(UINT4 *) (pair->strvalue));
						len = pair->size - 4;

						if (len > 0)
							sprintf(row_entries[c], "\'%d ", number);
						else
							sprintf(row_entries[c], "\'%d No class defined ", number);

						str_end = strlen(row_entries[c]);
						while (len)
						{
							sprintf(&(row_entries[c][str_end]), "%02x", *ptr);
							ptr++;
							len--;
							str_end += 2;
						}
						sprintf(&(row_entries[c][str_end]), "\'");
						break;
					}

				case PW_TYPE_STRING:
					row_entries[c] = malloc(strlen((u_char *) pair->strvalue) * 4);
					sprintf(row_entries[c], "\'");

					ptr = (u_char *) pair->strvalue;
					str_end = strlen(row_entries[c]);
					while (*ptr != '\0')
					{
						if (!(isprint(*ptr)))
						{
							sprintf(&(row_entries[c][str_end]), "\\%03o", *ptr);
							str_end += 4;
						}
						else
						{
							sprintf(&(row_entries[c][str_end]), "%c", *ptr);
							str_end++;
						}
						ptr++;
					}
					sprintf(&(row_entries[c][str_end]), "\'");
					break;

				case PW_TYPE_INTEGER:
					row_entries[c] = malloc(20);

					dval = dict_valget(pair->lvalue, pair->name);
					sprintf(row_entries[c], "%ld", pair->lvalue);
					break;

				case PW_TYPE_IPADDR:
					ipaddr2str(buffer, pair->lvalue);
					row_entries[c] = malloc(strlen(buffer) + 3);
					sprintf(row_entries[c], "\'%s\'", buffer);
					break;

				case PW_TYPE_DATE:
					row_entries[c] = malloc(strlen(buffer) + 3);
					sprintf(row_entries[c], "%lu", (time_t) pair->lvalue);
					break;

#if defined( BINARY_FILTERS )
				case PW_TYPE_FILTER_BINARY:
					{
						int                  ix;

						row_entries[c] = malloc((pair->lvalue * 3) + 3);

						sprintf(row_entries[c], "\'");
						str_end = strlen(row_entries[c]);

						for (ix = 0; ix < pair->lvalue; ix += 1)
						{
							sprintf(&(row_entries[c][str_end]), " %02x", (u_char) (pair->strvalue[ix]));
							str_end += 3;
						}
						sprintf(&(row_entries[c][str_end]), "\'");
					}
					break;
#endif /* BINARY_FILTERS */

				default:
/*             fprintf(fd, "Unknown type %d", pair->type); */
					break;
			}
		}
		pair = pair->next;
	}

	/* Set up the query buffer */
	query = malloc(strlen(SQL_DB_TABLE) + 35);
	sprintf(query, "INSERT INTO %s (Event_Recieved_Time", SQL_DB_TABLE);

	/* Print the field names into the query buffer */
	for (c = 0; c < SQL_NUM_ATTRIBS; c++)
	{
		if (row_entries[c])
		{
			query = realloc(query, strlen(query) + strlen(sql_field_names[c]) + 3);
			sprintf(&(query[strlen(query)]), ", %s", sql_field_names[c]);
		}
	}

	/* Print some interviening SQL stuff into the buffer */
	query = realloc(query, strlen(query) + 22);
	sprintf(&(query[strlen(query)]), ") VALUES (%lu", (unsigned long) time(NULL));

	/* Print the values into the buffer */
	for (c = 0; c < SQL_NUM_ATTRIBS; c++)
	{
		if (row_entries[c])
		{
			query = realloc(query, strlen(query) + strlen(row_entries[c]) + 3);
			sprintf(&(query[strlen(query)]), ", %s", row_entries[c]);

			/* Free the value strings as we go */
			free(row_entries[c]);
		}
	}

	/* Finish off the query string */
	query = realloc(query, strlen(query) + 2);
	sprintf(&(query[strlen(query)]), ")");

	SendQueryToDatabase(query);

	free(query);
}

/*************************************************************************
 *
 *	Function: SendQueryToDatabase
 *
 *	Purpose: Sends an SQL query to a database server
 *
 *************************************************************************/

void                 SendQueryToDatabase(char *query)
{
/* Connection descriptor */
	MYSQL                mysql;

/* Result from query */

	openlog("radiusd", LOG_PID | LOG_NDELAY, LOG_DAEMON);

	if (!mysql_connect(&mysql, SQL_DB_HOST, SQL_DB_USER, SQL_DB_PASS))
	{
		syslog(LOG_ERR, "DB Connect failed: %s", mysql_error(&mysql));
		return;
	}

	if (mysql_select_db(&mysql, SQL_DB_NAME))
	{
		syslog(LOG_ERR, "DB Select failed: %s", mysql_error(&mysql));
		return;
	}

	if (mysql_query(&mysql, query))
	{
		syslog(LOG_ERR, "DB Query failed: %s", mysql_error(&mysql));
		syslog(LOG_ERR, "Query: %s", query);
		return;
	}

	mysql_close(&mysql);
	
#ifdef DEBUG_SQL	
   syslog(LOG_NOTICE, "Query = %s", query);
#endif
}
